/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package com.sun.xml.ws.assembler;

import com.sun.xml.ws.api.ResourceLoader;
import com.sun.xml.ws.api.server.Container;
import com.sun.xml.ws.runtime.config.TubeFactoryConfig;
import com.sun.xml.ws.runtime.config.TubeFactoryList;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import junit.framework.TestCase;

/**
 *
 * @author Marek Potociar <marek.potociar at sun.com>
 */
public class MetroConfigLoaderTest extends TestCase {

    private static final class TestDataInfo {

        final String appConfigFileName;
        final int expectedTubes;

        public TestDataInfo(String appConfigFileName, int expectedTubes) {
            this.appConfigFileName = appConfigFileName;
            this.expectedTubes = expectedTubes;
        }
    }

    private static class TestDataLoader extends ResourceLoader {

        private final String resourceName;

        public TestDataLoader(String resourceName) {
            this.resourceName = UNIT_TEST_RESOURCE_ROOT + resourceName;
        }

        @Override
        public URL getResource(String resource) throws MalformedURLException {
            if ("metro.xml".equals(resource)) {
                return Thread.currentThread().getContextClassLoader().getResource(resourceName);
            } else {
                return Thread.currentThread().getContextClassLoader().getResource(resource);
            }
        }
    }
    private static final String UNIT_TEST_RESOURCE_ROOT = "metro-config/";
    private static final TestDataInfo[] APP_METRO_CONFIGS = new TestDataInfo[]{
        new TestDataInfo("metro.xml", 12),
        new TestDataInfo("metro-no-default.xml", 13),
        new TestDataInfo("metro-no-default-single-tube.xml", 1)
    };

    public MetroConfigLoaderTest(String testName) {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    private Container createMockupContainer(final ResourceLoader loader) {
        return new Container() {

            @Override
            public <T> T getSPI(Class<T> spiType) {
                if (ResourceLoader.class.isAssignableFrom(spiType)) {
                    return spiType.cast(loader);
                } else {
                    return null;
                }
            }
        };
    }

    /**
     * Test of getTubelineForEndpoint method, of class MetroConfigLoader.
     */
    public void testGetEndpointSideTubeFactoriesTest() throws URISyntaxException {
        TestDataInfo tdi = APP_METRO_CONFIGS[0];
        MetroConfigLoader configLoader = new MetroConfigLoader(createMockupContainer(new TestDataLoader(tdi.appConfigFileName)));

        TubeFactoryList result;
        result = configLoader.getEndpointSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/HttpPingPort)"));
        assertTrue(containsTubeFactoryConfig(result, "server"));
        assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());

        result = configLoader.getEndpointSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/JmsPingPort)"));
        assertTrue(containsTubeFactoryConfig(result, "default-server"));
        assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());

        result = configLoader.getEndpointSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/OtherPingPort)"));
        assertTrue(containsTubeFactoryConfig(result, "default-server"));
        assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());
    }

    /**
     * Test of getTubelineForEndpoint method, of class MetroConfigLoader.
     */
    public void testGetClientSideTubeFactoriesTest() throws URISyntaxException {
        TestDataInfo tdi = APP_METRO_CONFIGS[0];
        MetroConfigLoader configLoader = new MetroConfigLoader(createMockupContainer(new TestDataLoader(tdi.appConfigFileName)));

        TubeFactoryList result;
        result = configLoader.getClientSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/HttpPingPort)"));
        assertTrue(containsTubeFactoryConfig(result, "default-client"));
        assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());

        result = configLoader.getClientSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/JmsPingPort)"));
        assertTrue(containsTubeFactoryConfig(result, "client"));
        assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());

        result = configLoader.getClientSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/OtherPingPort)"));
        assertTrue(containsTubeFactoryConfig(result, "default-client"));
        assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());
    }

    /**
     * Test of getTubelineForEndpoint method, of class MetroConfigLoader - loading from default Metro config
     */
    public void testGetEndpointSideTubeFactoriesLoadFromDefaultConfig() throws URISyntaxException {
        for (int i = 1; i < APP_METRO_CONFIGS.length; i++) {
            TestDataInfo tdi = APP_METRO_CONFIGS[i];

            MetroConfigLoader configLoader = new MetroConfigLoader(createMockupContainer(new TestDataLoader(tdi.appConfigFileName)));

            TubeFactoryList result;
            result = configLoader.getEndpointSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/HttpPingPort)"));
            assertTrue(containsTubeFactoryConfig(result, "server"));
            assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());

            result = configLoader.getEndpointSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/JmsPingPort)"));
            assertFalse(result.getTubeFactoryConfigs().isEmpty());
            assertFalse(containsTubeFactoryConfig(result, "server"));

            result = configLoader.getEndpointSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/OtherPingPort)"));
            assertFalse(result.getTubeFactoryConfigs().isEmpty());
            assertFalse(containsTubeFactoryConfig(result, "server"));
        }
    }

    /**
     * Test of getTubelineForEndpoint method, of class MetroConfigLoader - loading from default Metro config
     */
    public void testGetClientSideTubeFactoriesLoadFromDefaultConfig() throws URISyntaxException {
        for (int i = 1; i < APP_METRO_CONFIGS.length; i++) {
            TestDataInfo tdi = APP_METRO_CONFIGS[i];

            MetroConfigLoader configLoader = new MetroConfigLoader(createMockupContainer(new TestDataLoader(tdi.appConfigFileName)));

            TubeFactoryList result;
            result = configLoader.getClientSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/HttpPingPort)"));
            assertFalse(result.getTubeFactoryConfigs().isEmpty());
            assertFalse(containsTubeFactoryConfig(result, "client"));

            result = configLoader.getClientSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/JmsPingPort)"));
            assertTrue(containsTubeFactoryConfig(result, "client"));
            assertEquals(tdi.expectedTubes, result.getTubeFactoryConfigs().size());

            result = configLoader.getClientSideTubeFactories(new URI("http://org.sample#wsdl11.port(PingService/OtherPingPort)"));
            assertFalse(result.getTubeFactoryConfigs().isEmpty());
            assertFalse(containsTubeFactoryConfig(result, "client"));
        }
    }

    private boolean containsTubeFactoryConfig(TubeFactoryList tubeList, String tubeFactoryName) {
        for (TubeFactoryConfig config : tubeList.getTubeFactoryConfigs()) {
            if (config.getClassName().equals(tubeFactoryName)) {
                return true;
            }
        }

        return false;
    }
}
